#include <vector>
#include <stdexcept>
#include "fonts.h"
#include "resolution.h"

std::vector<FT_Face> fonts;

void init_freetype()
{
	FT_Library library;
	FT_Face face;
	if (FT_Init_FreeType(&library))
		throw std::runtime_error("FreeType init failed.");

	// Georgia
	if (FT_New_Face(library, "/usr/share/fonts/truetype/msttcorefonts/Georgia.ttf", 0, &face))
		throw std::runtime_error("Face opening failed.");
	fonts.push_back(face);

	// FreeSerif
	if (FT_New_Face(library, "/usr/share/fonts/truetype/freefont/FreeSerif.ttf", 0, &face)) {
		std::fprintf(stderr, "Warning: Couldn't open FreeSerif, some glyphs might not be available\n");
	} else {
		fonts.push_back(face);
	}
	
	// Arial Unicode MS
	if (FT_New_Face(library, "arialuni.ttf", 0, &face)) {
		std::fprintf(stderr, "Warning: Couldn't open Arial Unicode MS, some glyphs might not be available\n");
	} else {
		fonts.push_back(face);
	}
}

// this should really be done somehow else :-)
static unsigned screen_width = SCREEN_WIDTH, screen_height = SCREEN_HEIGHT;
void set_screen_size(unsigned width, unsigned height)
{
	screen_width = width;
	screen_height = height;
}

unsigned my_draw_text(const widestring &str, unsigned char *buf, double size, int xpos, int ypos, int r, int g, int b)
{
	FT_GlyphSlot slot;
	int x = 0;
	
	for (std::vector<FT_Face>::const_iterator i = fonts.begin(); i != fonts.end(); ++i) {
		if (FT_Set_Char_Size(*i, 0, unsigned(size * 64.0), 96 * screen_width/LOGICAL_SCREEN_WIDTH, 96 * screen_height/LOGICAL_SCREEN_HEIGHT))
			throw std::runtime_error("Couldn't set font size");
	}

	// whoop :-P
	xpos = xpos * screen_width / LOGICAL_SCREEN_WIDTH;
	ypos = ypos * screen_height / LOGICAL_SCREEN_HEIGHT;

	for (widestring::const_iterator i = str.begin(); i != str.end(); ++i) {
		int glyph_index;
		for (std::vector<FT_Face>::const_iterator j = fonts.begin(); j != fonts.end(); ++j) {
			glyph_index = FT_Get_Char_Index(*j, *i);
			if (glyph_index == 0)
				continue;

#if SCREEN_LCD
			if (FT_Load_Glyph(*j, glyph_index, FT_LOAD_RENDER | FT_LOAD_TARGET_LCD))
#else
			if (FT_Load_Glyph(*j, glyph_index, FT_LOAD_RENDER))
#endif
				throw std::runtime_error("Couldn't load glyph");
			slot = (*j)->glyph;
			break;
		}
		if (glyph_index == 0) {
			std::fprintf(stderr, "Warning: Could not find a glyph in any font for U+%x, ignoring\n", *i);
			continue;
		}

		if (buf != NULL) {
			int y;
			FT_Bitmap *bm = &(slot->bitmap);
			for (y = 0; y < bm->rows; y++) {
				int xx;
				int dsty = ypos - slot->bitmap_top + y;
				if (dsty < 0 || dsty > signed(screen_height-1)) continue;

				unsigned char *dst = buf + dsty * screen_width*4 + (x + xpos + slot->bitmap_left)*4;
#if SCREEN_LCD
				unsigned char *src = bm->buffer + y * bm->pitch;
				int width = (x + xpos + slot->bitmap_left + bm->width/3 >= signed(screen_width)) ? ((screen_width-1) - x - xpos - slot->bitmap_left) : bm->width/3;
#else
				unsigned char *src = bm->buffer + y * bm->pitch;
				int width = (x + xpos + slot->bitmap_left + bm->width >= signed(screen_width)) ? ((screen_width-1) - x - xpos - slot->bitmap_left) : bm->width;
#endif

#if SCREEN_LCD
				for (xx = 0; xx < width; xx++) {
					*dst = (*dst * (256-src[2]) + b * src[2]) >> 8;
					++dst;
					*dst = (*dst * (256-src[1]) + g * src[1]) >> 8;
					++dst;
					*dst = (*dst * (256-src[0]) + r * src[0]) >> 8;
					++dst;
					*dst++ = 0;

					src += 3;
				}
#else
				for (xx = 0; xx < width; xx++) {
					*dst = (*dst * (256-*src) + b * *src) >> 8;
					++dst;
					*dst = (*dst * (256-*src) + g * *src) >> 8;
					++dst;
					*dst = (*dst * (256-*src) + r * *src) >> 8;
					++dst;
					*dst++ = 0;
					++src;
				}
#endif
			}
		}

		x += slot->advance.x >> 6;
	}

	return x * LOGICAL_SCREEN_WIDTH / screen_width;
}

void my_draw_text_deferred(std::vector<TextDefer> &td, const widestring &str, double size, int xpos, int ypos)
{
	TextDefer newtd;
	newtd.str = str;
	newtd.size = size;
	newtd.xpos = xpos;
	newtd.ypos = ypos;
	td.push_back(newtd);
}

void draw_all_deferred_text(unsigned char *buf, std::vector<TextDefer> &current, std::vector<TextDefer> &old)
{
	for (unsigned i = 0; i < current.size(); ++i) {
		int r, g, b;
		if (i < old.size() && current[i].str != old[i].str) {
			// changed text
			r = 255;
			g = 0;
			b = 0;
		} else {
			r = g = b = 255;
		}	
		
		my_draw_text(current[i].str, buf, current[i].size, current[i].xpos, current[i].ypos, r, g, b);
	}
}

